import axios from 'axios';
import fs from 'fs';
import path from 'path';
import { execSync } from 'child_process';

/**
 * Options for the API generator
 */
export interface ApiGeneratorOptions {
  /**
   * URL to the OpenAPI schema
   */
  schemaUrl: string;
  
  /**
   * Output directory for the generated API client
   */
  outputDir: string;
  
  /**
   * Whether to skip TLS verification (useful for local development)
   * @default false
   */
  skipTlsVerify?: boolean;
  
  /**
   * Whether to run post-processing on the generated files
   * @default true
   */
  runPostProcessing?: boolean;
}

/**
 * Downloads the OpenAPI schema from the specified URL
 */
export async function downloadSchema(options: ApiGeneratorOptions): Promise<string> {
  const { schemaUrl, outputDir, skipTlsVerify = false } = options;
  const outPath = path.join(outputDir, 'open-api.json');
  
  console.log(`Downloading schema from ${schemaUrl}...`);
  
  try {
    // Create output directory if it doesn't exist
    fs.mkdirSync(path.dirname(outPath), { recursive: true });
    
    // Download schema
    const axiosConfig = skipTlsVerify ? { httpsAgent: new (require('https').Agent)({ rejectUnauthorized: false }) } : {};
    const response = await axios.get(schemaUrl, axiosConfig);
    
    // Process schema
    const schema = response.data;
    
    // Remove servers array to avoid hardcoded URLs
    if (schema.servers) {
      delete schema.servers;
    }
    
    // Write schema to file
    fs.writeFileSync(outPath, JSON.stringify(schema, null, 2));
    
    // Check for duplicate DTOs
    console.log('Checking for duplicate DTOs in schema...');
    const schemaContent = fs.readFileSync(outPath, 'utf8');
    const duplicateDtoRefRegex = /"\$ref":\s*"#\/components\/schemas\/\w+2"/;
    
    if (duplicateDtoRefRegex.test(schemaContent)) {
      throw new Error(
        "The schema contains duplicate DTOs (pattern: $ref: '#/components/schemas/...Name2'). " +
        "This is a transient error, usually due to a bug in OpenAPI generators. " +
        "Try restarting the API server."
      );
    }
    
    console.log('Schema downloaded and processed successfully.');
    return outPath;
  } catch (error) {
    if (axios.isAxiosError(error)) {
      throw new Error(`Failed to download schema: ${error.message}`);
    }
    throw error;
  }
}

/**
 * Generates the API client using openapi-typescript-codegen
 */
export function generateClient(schemaPath: string, outputDir: string): void {
  console.log('Generating API client...');
  
  const codegenCmd = [
    'npx openapi-typescript-codegen',
    '--input', schemaPath,
    '--output', outputDir,
    '--client', 'axios',
    '--useOptions'
  ].join(' ');
  
  execSync(codegenCmd, { stdio: 'inherit' });
  
  // Add auto-generated comment to each file
  const files = fs.readdirSync(outputDir);
  for (const file of files) {
    const filePath = path.join(outputDir, file);
    if (file.startsWith('openapi') && file.endsWith('.ts')) {
      const content = fs.readFileSync(filePath, 'utf8');
      const modifiedContent = `/* AUTOGENERATED - DO NOT EDIT */\n\n${content}`;
      fs.writeFileSync(filePath, modifiedContent);
    }
  }
  
  console.log('API client generated successfully.');
}

/**
 * Post-processes the generated API client files
 */
export function postProcessFiles(outputDir: string): void {
  console.log('Post-processing generated files...');
  
  const modelsDir = path.join(outputDir, 'models');
  const servicesDir = path.join(outputDir, 'services');
  
  const directoriesToProcess = [
    modelsDir,
    servicesDir
  ];
  
  function processFile(filePath: string): void {
    let fileContent = fs.readFileSync(filePath, 'utf8');
    
    // Remove import statements for 'void'
    fileContent = fileContent.replace(/import type \{ void \} from '.*';/g, '');
    
    // Replace usages of 'void' as a type with 'any'
    fileContent = fileContent.replace(/: void,/g, ': any,');
    fileContent = fileContent.replace(/: void \{/g, ': any {');
    fileContent = fileContent.replace(/: void;/g, ': any;');
    
    // Detect and replace self-referencing DTOs with 'any'
    const typeDefinitionRegex = /export type (\w+) = (.*?);/gs;
    let match;
    while ((match = typeDefinitionRegex.exec(fileContent)) !== null) {
      const typeName = match[1];
      const typeDefinition = match[2];
      if (typeDefinition.includes(typeName)) {
        const newTypeDefinition = typeDefinition.replace(new RegExp(`\\b${typeName}\\b`, 'g'), 'any');
        fileContent = fileContent.replace(typeDefinition, newTypeDefinition);
      }
    }
    
    fs.writeFileSync(filePath, fileContent, 'utf8');
  }
  
  function processDirectory(dirPath: string): void {
    if (!fs.existsSync(dirPath)) {
      return;
    }
    
    const files = fs.readdirSync(dirPath);
    files.forEach(file => {
      const filePath = path.join(dirPath, file);
      const stat = fs.statSync(filePath);
      if (stat.isDirectory()) {
        processDirectory(filePath);
      } else if (path.extname(file) === '.ts') {
        processFile(filePath);
      }
    });
  }
  
  directoriesToProcess.forEach(dir => {
    processDirectory(dir);
  });
  
  console.log('Post-processing complete.');
}

/**
 * Main function to generate the API client
 */
export async function generateApiClient(options: ApiGeneratorOptions): Promise<void> {
  try {
    const schemaPath = await downloadSchema(options);
    generateClient(schemaPath, options.outputDir);
    
    if (options.runPostProcessing !== false) {
      postProcessFiles(options.outputDir);
    }
    
    console.log(`API client successfully generated in ${options.outputDir}`);
  } catch (error) {
    console.error('Error generating API client:', error);
    throw error;
  }
}
